// EODatabase.h
// Copyright (c) 1994, NeXT Software, Inc. All rights reserved.

#import <EOControl/EOControl.h>
#import <EOAccess/EODefines.h>

@class EOEditingContext;
@class EOGlobalID;
@class EOEntity;
@class EODatabaseContext;
@class EOModel;
@class EOAdaptor;

// Exceptions raised by EODatabase layer when errors occur during DBMS interactions.
EOACCESS_EXTERN NSString *EOGeneralDatabaseException;

/* An EODatabase represents a database server. It contains a list of databaseContexts that are connected to this server, a list of models which describe the server's schema, an adaptor which is is capable of communicating with the server and a set of snapshots representing the state of all objects stored in this server */

@interface EODatabase:NSObject
{
    NSMutableArray *_registeredContexts;
    NSMutableDictionary *_snapshots;
    NSMutableArray *_models;
    void *_reserved;  // WARNING: adding ivars will break binary compatability
}

- initWithAdaptor:(EOAdaptor *)adaptor;
    // Initializes a newly allocated EODatabase with adaptor as its adaptor.
    // Returns self.  You must never associate more than one EODatabase with
    // a given adaptor instance.

- initWithModel:(EOModel *)model;
    // Calls [EOAdaptor adaptorWithModel:] and then [self initWithAdaptor:]
    // and [self addModel:]

- (NSArray *)registeredContexts;
    // Returns an array containing all contexts that have been registered for
    // use with this database;

- (void)registerContext:(EODatabaseContext *)context;
- (void)unregisterContext:(EODatabaseContext *)context;
    // Registers or unregisters a channel with a EODatabaseContext. Normally,
    // these methods are called by [EODatabaseContext initWithDatabaseContext:]
    // and [EODatabaseContext dealloc] respectively.

- (EOAdaptor *)adaptor;
    // Returns the EOAdaptor instance used by the EODatabase for communication
    // with the database server.

// Model list support
- (void)addModel:(EOModel *)model;
- (void)removeModel:(EOModel *)model;
- (BOOL)addModelIfCompatible:(EOModel *)model;
    // Returns YES if model already added.  If not, and this model specifies
    // the same adaptor name as the database's adaptor and the adaptor responds
    // YES to canServiceModel:, adds it and returns YES.
    // Otherwise returns NO.

- (NSArray *)models;
    // Returns the list of models held by this EODatabase.

- (EOEntity *)entityNamed:(NSString *)entityName;
    // Forwards the entityNamed: method to each model in the model list and returns
    // the first entity found with entityName. Returns nil if non is found.

- (EOEntity *)entityForObject:(id)object;
    // Forwards the entityForObject: method to each model in the model list and returns
    // the first entity found. Returns nil if non is found.

- (NSArray *)resultCacheForEntityNamed:(NSString *)name;
- (void)setResultCache:(NSArray *)cache forEntityNamed:(NSString *)name;
- (void)invalidateResultCacheForEntityNamed:(NSString *)name;
- (void)invalidateResultCache;
    // Sets/Returns an array of gid's corresponding to every object in the underlying
    // repository. This cache can be used by the database context to evaluate
    // qualifiers in memory rather than in the database.
@end


// The EOUniquing category defines methods for recording and forgetting snapshots
// of an object based on the object's globalID

@interface EODatabase (EOUniquing)

+ (void)setDatabaseDoesReleaseUnreferencedSnapshots:(BOOL)yn;
    // Turn on or off the releasing of unreferenced snapshots by the EODatabase.
    // Default is YES (the database does release snapshots).

+ (BOOL)databaseDoesReleaseUnreferencedSnapshots;

- (void)recordSnapshot:(NSDictionary *)snapshot forGlobalID:(EOGlobalID *)gid;
    // Records the snapshot for the GlobalID.

- (NSDictionary *)snapshotForGlobalID:(EOGlobalID *)gid;
    // Returns the snapshot assocated with the GlobalID if there is one;
    // else returns nil.

- (void)recordSnapshot:(NSArray *)gids forSourceGlobalID:(EOGlobalID *)gid relationshipName:(NSString *)name;
    // Record the to many relationship (represented by an array of global IDs in the snapshot table.

- (NSArray *)snapshotForSourceGlobalID:(EOGlobalID *)gid relationshipName:(NSString *)name;
    // Return the to many snapshot for this globalID/relationship name.

- (void)forgetSnapshotForGlobalID:(EOGlobalID *)gid;
    // Remove the snapshot from snapshot table

- (void)forgetSnapshotsForGlobalIDs:(NSArray *)array;
    // Removes each snapshot from snapshot table

- (void)forgetAllSnapshots;
    // Removes all snapshots from the snapshot table.

- (void)recordSnapshots:(NSDictionary *)snapshots;
    // record a bunch of snapshots from a dictionary keyed by Global ID

- (void)recordToManySnapshots:(NSDictionary *)snapshots;
    // record a bunch of snapshots from a dictionary keyed by Global ID

- (NSDictionary *)snapshots;
    // returns dictionary of snapshots keyed by GlobalID

- (void)incrementSnapshotCountForGlobalID:(EOGlobalID *)gid;
    // If snapshot releasing is turned on, increment the reference count 
    // for the shared snapshot keyed by Global ID.

- (void)decrementSnapshotCountForGlobalID:(EOGlobalID *)gid;
    // If snapshot releasing is turned on, decrement the reference count 
    // for shared snapshot keyed by Global ID;  if no more objects
    // refer to the snapshot, remove it from the snapshot table.

- (int)snapshotRetainCountForGlobalID:(EOGlobalID *)gid;
    // If snapshot releasing is turned on, return the retain count for 
    // the shared snapshot keyed by the given Global ID; otherwise return -1;
@end

